#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/stat.h>
#include <signal.h>
#include <linux/soundcard.h>
#include <errno.h>

#include "play.h"
extern int noerrs;
unsigned int sigintflag=0;

int header_check(wave_header *head) {
	unsigned p;
	
	p = head->length - head->data_length;

	if ( (p != 36) && (p != 37) && !noerrs ) {
		err(0,"Warning: Wrong length difference."); 
	}

	if (!noerrs) {
		if ( strncmp(head->main_chunk,"RIFF",4) ) {
			err(0,"Not in RIFF format.");
			return 0;
		}

		if ( strncmp(head->chunk_type,"WAVE",4) ) {
			err(0,"Not a WAVE type.");
			return 0;
		}
		
		if ( strncmp(head->sub_chunk,"fmt ",4) ) {
			err(0,"Not a WAVE format.");
			return 0;
		}

		if ( (head->sample_fq * head->byte_p_spl) != head->byte_p_sec ) {
			err(0,"Incorrect sampling rate.");
			return 0;
		}
       }	
       return 1;
}


/* Signal handler for SIGINT, we want to close the wav file correctly
 * and not a corrupted wav-file
 */

void signalhandler(int signr) {
	sigintflag=1;
	return;
}


/*********************************************************************
 * Record or Play samples :
 *
 * mode = PLAY / RECORD
 * data = command_line options
 *********************************************************************/
void recplay(char *filename,int mode,header_data data,int ipc) {
	int tmp;
	int count;			/* Total # of bytes */
	int cnt;			/* bytes for this I/O */
	int n;				/* Bytes read/written */
	int omode = (mode == PLAY) ? O_WRONLY : O_RDONLY;	
	int um = 0666;			/* Current umask() value */
	int fd = -1;			/* read/write file descriptor */
	int audio = -1;			/* /dev/dsp audio file descriptor */
	int abuf_size;			/* audio buffer size for I/O */
	char *iobuf;			/* allocated I/O buffer */
	char tbuf[80];			/* Temporary buffer */
	wave_header whdr;		/* header to/from file */

	/*************************************************************
	 * Find out what are umask setting is :
	 *************************************************************/

	um = umask(0666);
	umask(um);		/* Restore umask */

	/*************************************************************
	 * Lock the device :
	 *************************************************************/

	if ( ipc >= 0 ) {
		if ( LockDSP(ipc,mode==PLAY?0:1) ) {
			ipc = -1;		/* Don't unlock now */
			err(1,mode==PLAY
				? "Locking audio device for play."
				: "Locking audio device for record.");
			goto errxit;
		}
	}

	if ( mode == PLAY ) {
		/*****************************************************
		 * Play a sample :
		 *****************************************************/

		if ( filename == NULL ) {
			fd = 0;
			filename = "stdin";
		} else if ( (fd = open(filename,O_RDONLY,0)) < 0 ) {
			sprintf(tbuf,"Error opening %s",filename);
			err(1,tbuf);
			return;
		}

		if ( (n = read(fd,&whdr,sizeof(whdr))) < 0 ) {
			sprintf(tbuf,"Reading %s.",filename);
			err(1,tbuf);
			goto errxit;

		} else if ( n == 0 )
			goto xit;	/* Perhaps its /dev/null */

		else if ( n < sizeof(whdr) ) {
			err(0,"Short header: can't be wav file.");
			goto errxit;
		}
		
		if ( !header_check(&whdr) )
			return; 
		/* Set play parameters */
		if (!noerrs) {
			if ( !data.force_stereo )
				data.stereo = (whdr.modus == 2);

			if ( !data.force_speed )
				data.speed = whdr.sample_fq;

			if ( !data.force_sample_size ) 
				data.sample_size = whdr.bit_p_spl;
		} else { /* ignore header, use default values */
			if ( !data.force_stereo )
				data.stereo = 0;

			if ( !data.force_speed )
				data.speed = 22050;

			if ( !data.force_sample_size ) 
				data.sample_size = 8;
		}	
			
		/*****************************************************
		 * set upper border accordingly
		 *****************************************************/
			
		if ( data.time_limit == 0.0 ) 
			data.upper_border = whdr.data_length;

		else	{
			if (!data.force_speed)
				data.upper_border = data.time_limit * whdr.sample_fq;
			else	data.upper_border = data.time_limit * data.speed;
		
			if ( data.stereo )
				data.upper_border *= 2;
			if ( data.sample_size != 8 )
				data.upper_border *= 2;
		}
	} else	{				/* mode = RECORD */
		/*****************************************************
		 * Record a sample :
		 *****************************************************/

		if ( filename == NULL ) {
			fd = 1;
			filename = "stdout";
		} else	{
			remove(filename);
			if ( (fd = open(filename,O_WRONLY|O_CREAT,0666&~um)) < 0 ) {
				sprintf(tbuf,"Opening %s for write.",filename);
				err(1,tbuf);
				return;
			}
		}

		/*****************************************************
		 * Create header :
		 *****************************************************/

		strncpy(whdr.main_chunk,"RIFF",4);
		strncpy(whdr.chunk_type,"WAVE",4);
		strncpy(whdr.sub_chunk,"fmt ",4);
		strncpy(whdr.data_chunk,"data",4);

		whdr.length_chunk = 16;
		whdr.format = 1;
		whdr.modus = data.stereo ? 2 : 1;		
		whdr.sample_fq = data.speed;

		if ( !data.stereo ) {
			/*********************************************
			 * Mono :
			 *********************************************/
			if ( data.sample_size == 8 )
				whdr.byte_p_spl = 1;
			else 	whdr.byte_p_spl = 2;
		} else	{
			/*********************************************
			 * Stereo :
			 *********************************************/
			if ( data.sample_size == 8 )
				whdr.byte_p_spl = 2;
			else	whdr.byte_p_spl = 4;
		}

		whdr.byte_p_sec = data.speed * whdr.byte_p_spl;
		whdr.data_length = whdr.byte_p_sec * data.time_limit;

		whdr.bit_p_spl = data.sample_size;
		whdr.length = whdr.data_length + 36;
		data.upper_border = data.time_limit * data.speed
			* whdr.byte_p_spl;
		data.lower_border = 0;

		if ( write(fd,&whdr,sizeof(whdr)) != sizeof(whdr) ) {
			err(1,"Writing WAVE header.");
		   	return;
		}
	}

	count = data.upper_border - data.lower_border;

	if ( !data.quiet_mode ) {
		if ( mode == PLAY )
			fprintf(stderr,"Playing: %s",filename);
		else	fprintf(stderr,"Recording: %s",filename);

		if ( data.time_limit == 0.0 ) 
			fprintf(stderr," (in full)\n");
		else	fprintf(stderr," (%5f secs)\n",data.time_limit);

		if (noerrs)
			fprintf(stderr, "using default values:\n");
		if ( data.stereo ) 
			fprintf(stderr,"Stereo");
		else	fprintf(stderr,"Mono");
		fprintf(stderr,", rate %u Hz, %u bits.\n", 
			data.speed, data.sample_size);
	}

	if ( (audio = open(AUDIODEV,omode,0)) < 0 ) {
		err(1,"Error opening audio-device");
		return;
	}

	if ( ioctl(audio,SNDCTL_DSP_GETBLKSIZE,&abuf_size) < 0 ) {
		err(1,"Obtaining dsp buffer size.");
		goto errxit;
	}
#if 1
	if ( abuf_size < 1 || abuf_size > 65536 ) {
		sprintf(tbuf,"Invalid audio buffer size (%d bytes).",(int)abuf_size);
		err(1,tbuf);
		goto errxit;
	} 
#else
	abuf_size=1;
#endif /* 0 */
	if ( (iobuf = malloc(abuf_size)) == NULL ) {
		err(1,"Unable to allocate audio buffer.");
		goto errxit;
	}

        tmp = data.sample_size;
        if ( ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &data.sample_size) < 0 ) {
        	err(1,"Unable to set audio sample size.");
		goto errxit;
	}

	if ( tmp != data.sample_size ) { /* Is this test needed? */
		err(0,"Unable to set sample size");
		goto errxit;
	}

	if ( ioctl(audio, SNDCTL_DSP_STEREO, &data.stereo) < 0 ) {
		err(1,"Unable to set stereo/mono.");
		goto errxit;
	}		
      
	if ( ioctl(audio,SNDCTL_DSP_SPEED,&data.speed) < 0 ) {
		err(1,"Unable to set audio sampling rate.");
		goto errxit;
	}

	if ( mode == PLAY ) {
		/*****************************************************
		 * Play :
		 *****************************************************/
		if (noerrs)
			count=0x7fffffff;
		while( count > 0 ) {
			cnt = count > abuf_size ? abuf_size : count;

			if ( (n = read(fd,iobuf,cnt)) < 0 ) {
				err(1,"Reading audio sample.");
				goto errxit;
			}
			if (! data.quiet_mode )
				drawbar(iobuf, n, data);
	
			if ( n == 0 ) {
				if (!noerrs) 
					err(0,"Unexpected EOF reading samples.");
				fflush(NULL);
				goto errxit;
			}

			if ( write(audio,iobuf,n) != n ) {
				err(1,"Writing audio device.");
				goto errxit;
			}

			count -= n;
		}
			
	} else	{
		/*****************************************************
		 * Record :
		 * Install a signal Handler for sigint, we want 
		 * to close the WAV File correctly 
		 *****************************************************/
		signal(SIGINT, signalhandler);
		while ( (count > 0) && (sigintflag==0) ) {
			cnt =  count > abuf_size ? abuf_size : count;

			if ( (n = read(audio,iobuf,cnt)) < 0 ) {
				if ((errno!=EINTR) && (sigintflag==0)) {
					err(1,"Reading from audio device.");
					goto errxit;
				}
				if (sigintflag)
					break;
			
			}
			if (! data.quiet_mode )
				drawbar(iobuf, n, data);
			if ( write(fd,iobuf,n) != n ) {
				if ((errno!=EINTR) && (sigintflag==0)) {
					err(1,"Writing to output.");
					goto errxit;
				}
			}
			
			count -= n;
		}
	/* If we have been interrupted by SIGINT, manipulate wav header
	 */
	if (sigintflag == 1) {
		lseek(fd, 0, SEEK_SET);
		whdr.data_length=(data.upper_border-data.lower_border)-count;
		whdr.length=sizeof(whdr)+whdr.data_length;
		write (fd, &whdr, sizeof(whdr));
		}
	}

	/*************************************************************
	 * Normal Exit :
	 *************************************************************/

xit:	close(fd);
	close(audio);
	if ( ipc >= 0 ) {
		/*****************************************************
		 * Unlock so other processes can use it now :
		 *****************************************************/
		if ( UnlockDSP(ipc,mode==PLAY?0:1) ) {
			err(1,mode==PLAY
				? "Unlocking for audio play."
				: "Unlocking for audio record.");
		}
	}
	return;

	/*************************************************************
	 * Error Exit :
	 *************************************************************/

errxit:	if ( fd > 0 )
		close(fd);
	close(audio);
	if ( ipc >= 0 )
		UnlockDSP(ipc,mode==PLAY?0:1);	/* Best effort basis */
	exit(1);
}

/* Analyse the buffer and draw a bar showing the "aussteuerung" */

int drawbar (unsigned char * buffer, int size, header_data data) {
			
	int numbars = 1;		/* Number of bars */
	int factor = 1;			/* "Samplesize" one or two bytes? */
	int maxbarval0;		/* The maximum value of the two bars is calculated here */
	int maxbarval1;
	int barlength = 50;		/* The length of a bar */
	char s[100];
	double mult;
	int samplesize;
	int samplerate;
	int mode;
	int i;

	samplesize = data.sample_size;
	samplerate = data.speed;
	mode       = data.stereo;

	if (mode == 1) {
		numbars = 2;
	} else {
		numbars = 1;
	}
	if (samplesize > 8) {
		factor = 2;
		mult = (double)barlength/32768; 
	} else {
		factor = 1;
		mult = (double)barlength/128;
	}
	maxbarval0=0; maxbarval1=0;
	
	/* Check the values */
	for (i=0; i <= size ; i = i + (factor * numbars)) {
//		printf("%x ", abs((signed short int)(buffer[i]+256*buffer[i+1])));
		if ( factor == 1) {
			if ( buffer[i]-127 > maxbarval0) {
				maxbarval0 = buffer[i]-127;
			}
			if ( buffer[i+1]-127 > maxbarval1)
				maxbarval1 = buffer[i+1]-127;
		} else {	
			if ( abs((signed short int)(buffer[i]+256*buffer[i+1])) > maxbarval0)
				maxbarval0 = abs((signed short int)(buffer[i]+256*buffer[i+1]));
			if ( abs((signed short int)(buffer[i+2]+256*buffer[i+3])) > maxbarval1) {
				maxbarval1 = abs((signed short int)(buffer[i+2]+256*buffer[i+3]));
			}
		}
	}
	/*	printf("maxbarval0: %d, maxbarval1: %d\n", maxbarval[0], maxbarval[1]); */
	/*	printf("maxbarval0: %d, maxbarval1: %d\n", maxbarval0, maxbarval1); */
	for (i=0; i < (int)maxbarval1 * mult ; i++) {
		printf(">");
	} 
	for (i; i<barlength; i++) {
		printf("-");
	}	
	printf("|\n");
	if (numbars == 2) {
		for (i=0; i< (int)maxbarval0*mult; i++) {
			printf(">");
		}
		for (i; i<barlength; i++) {
			printf("-");
		}
		printf("|\n");
		printf("\e[A\e[A");
	} else {
		printf("\e[A");
	}
//	printf("\e]B");
	return 0;	
}
